/* kvlist.c
**
** Bilde aus dem abstrakten Datentyp `BBStree eine Datenstruktur zur
** Darstellung assoziativer Text-Arrays. Die in den Baum eingehngte
** Datenstruktur ist hierbei eine Liste von Zeichenketten, die smtliche
** Eintrge mit dem gleichen Schlssel reprsentiert.
** 
**
*/
#include <stdio.h>
#include <string.h>

#include "tools.h"
#include "mrmacs.h"

#include "kvlist.h"

typedef struct {
    kvitems_t first, last;
    int n;
    char *key;
} hlfifo_t;

/* Initialisiere ein assoziatives Array */
void kvlist_init (kvlist_t *al)
{
    al->t = (BBStree_t) NULL;
}

/* Vergleichsfunktion zur Suche eines Schlssels in einem assoziativen Array */
static int hlfifokey_cmp (char *key, hlfifo_t *el)
{
    return strcmp (key, el->key);
}

/* Vergleichsfunktion zur Eintragung eines neuen Elementes in das assoziative
** Array.
*/
static int hlfifo_cmp (hlfifo_t *l, hlfifo_t *r)
{
    return strcmp (l->key, r->key);
}

/* Fge ein neues Schlssel/Wert-Paar ein, oder hnge den Wert an die Liste
** von bereits zu diesem existierenden Werten am Ende an.
*/
int kvlist_insert (kvlist_t *al, char *key, char *value)
{
    hlfifo_t *x = (hlfifo_t *) BBStree_get (al->t, key,
					    (compare_func) hlfifokey_cmp);
    kvitems_t v;
    /* Zuerst mu ein Eintrag fr das neue Element generiert werden. */
    v = (kvitems_t) galloc (__FILE__, __LINE__,
			    sizeof(struct kvitems_s) + strlen (value) + 1);
    v->next = (kvitems_t) NULL;
    v->value = (char *) v + sizeof(struct kvitems_s);
    strcpy (v->value, value);
    ifnull (x) {
	/* Wenn der Schlssel bisher noch nicht vorkam, dann wird ebenfalls ein
	** Listenkopf erzeugt, in den das eben erzeugte Element eingehngt
	** wird; dieser Listenkopf wird anschlieend als neuer Knoten in den
	** Baum eingehngt.
	*/
	x = (hlfifo_t *) galloc (__FILE__, __LINE__,
				 sizeof (hlfifo_t) + strlen (key) + 1);
	x->first = x->last = v;
	x->n = 1;
	x->key = (char *) x + sizeof(hlfifo_t);
	strcpy (x->key, key);
	(void) BBStree_insert (&(al->t), x, (compare_func) hlfifo_cmp);
    } else {
	/* Andernfalls wird das neue Element einfach hinten an die Liste, die
	** an der entsprechenden Stelle steht, eingehngt.
	*/
	x->last->next = v;
	x->last = v;
	++x->n;
    }
    return x->n;
}

/* Suche nach einem Schlssel und gib die Liste der zugehrigen Werte zurck
** - oder NULL, falls der Schlssel nicht gefunden wurde.
*/
kvitems_t kvlist_lookup (kvlist_t *al, char *key)
{
    hlfifo_t *x = (hlfifo_t *) BBStree_get (al->t, key,
					    (compare_func) hlfifokey_cmp);
    ifnull (x) { return (kvitems_t) NULL; }
    return x->first;
}

/* Hilfsfunktion zur Ermittelung smtlicher Schlssel des assoziativer Arrays */
static int keys_traverse (hlfifo_t *trdata, hlfifo_t *list)
{
    kvitems_t v;
    v = (kvitems_t) galloc (__FILE__, __LINE__,
			    sizeof(struct kvitems_s) + strlen(trdata->key) + 1);
    v->next = (kvitems_t) NULL;
    v->value = (char *) v + sizeof(struct kvitems_s);
    strcpy (v->value, trdata->key);
    ifnull (list->first) {
	list->first = list->last = v;
	list->n = 1;
    } else {
	list->last->next = v;
	list->last = v;
	++list->n;
    }
    return 0;
}

/* Gib eine Liste smtlicher existierender Schlssel zurck - oder NULL, falls
** das Array leer ist.
*/
kvitems_t kvlist_keys (kvlist_t *al)
{
    hlfifo_t fifo;
    fifo.first = fifo.last = (kvitems_t) NULL;
    fifo.n = 0;
    (void) BBStree_lotraverse (al->t, (trav_func) keys_traverse, &fifo);
    return fifo.first;
}

static void hlfifo_free (hlfifo_t *fifo)
{
    kvitems_t p, q;
    for (p = fifo->first; noNULL (p); p = q) {
	q = p->next;
	gfree (p);
    }
    gfree (fifo);
}

int kvlist_delete (kvlist_t *al, char *key)
{
    return BBStree_delete (&al->t, key, (compare_func) hlfifokey_cmp,
					(release_func) hlfifo_free);
}

void kvlist_remove (kvlist_t *al)
{
    BBStree_clear (&al->t, (release_func) hlfifo_free);
    al->t = (BBStree_t) NULL;
}
